@@ -4,12 +4,14 @@ Verilog Abstract Syntax Tree Components
 
 from __future__ import annotations
 
-import itertools
-from typing import Optional
+from typing import Iterable, Optional
 
-from ... import ir
-from ...utils.assertions import assert_typed_dict, get_typed, get_typed_list
-from ...utils.string import ImplementsToLines, Indent, Lines
+from python2verilog import ir
+from python2verilog.utils import env
+from python2verilog.utils.generics import GenericRepr
+from python2verilog.utils.lines import ImplementsToLines, Indent, Lines
+from python2verilog.utils.mit_license import get_mit_license
+from python2verilog.utils.typed import guard, typed, typed_list, typed_strict
 
 
 class AtPosedge(ir.Expression):
@@ -34,15 +36,15 @@ class AtNegedge(ir.Expression):
         super().__init__(f"@(negedge {condition.to_string()})")
 
 
-class Statement(ImplementsToLines):
+class Statement(ImplementsToLines, GenericRepr):
     """
     Represents a statement in verilog (i.e. a line or a block)
     If used directly, it is treated as a string literal
     """
 
     def __init__(self, literal: str = "", comment: str = ""):
-        self.literal = get_typed(literal, str)
-        self.comment = get_typed(comment, str)
+        self.literal = typed(literal, str)
+        self.comment = typed(comment, str)
 
     def to_lines(self):
         """
@@ -67,6 +69,7 @@ class Statement(ImplementsToLines):
         // <comment>
         Separated by newlines
         """
+        assert guard(self.comment, str)
         actual_lines = self.comment.split("\n")
         lines = Lines()
         for line in actual_lines:
@@ -79,9 +82,37 @@ class LocalParam(Statement):
     localparam <name> = <value>;
     """
 
-    def __init__(self, name: ir.Expression, value: ir.UInt, *args, **kwargs):
+    def __init__(self, name: str, value: ir.UInt, *args, **kwargs):
         assert isinstance(value, ir.UInt)
-        super().__init__(f"localparam {name} = {value};", *args, **kwargs)
+        super().__init__(f"localparam {name} = {value.verilog()};", *args, **kwargs)
+
+
+class TypeDef(Statement):
+    """
+    typedef enum
+    {
+        <val0>, <val1>, ...
+    } _state_t;
+    """
+
+    def __init__(self, name: str, values: list[str]):
+        self.name = typed(name, str)
+        self.values = typed_list(values, str)
+        super().__init__()
+
+    def to_lines(self):
+        lines = Lines("typedef enum")
+        lines += "{"
+        values = Lines()
+        for value in self.values[:-1]:
+            values += f"{value},"
+        values += f"{self.values[-1]}"
+        lines.concat(values, indent=1)
+        lines += f"}} {self.name};"
+
+        # condense
+        lines = Lines(lines.to_string().replace("\n", "").replace("  ", ""))
+        return lines
 
 
 class AtPosedgeStatement(Statement):
@@ -114,7 +145,7 @@ class Instantiation(Statement):
         self,
         module_name: str,
         given_name: str,
-        port_connections: dict[ir.Expression | str, ir.Expression | str],
+        port_connections: dict[str, str],
         *args,
         **kwargs,
     ):
@@ -156,14 +187,16 @@ class Module(ImplementsToLines):
         outputs: list[str],
         body: Optional[list[Statement]] = None,
         is_not_testbench=True,
-        localparams: Optional[dict[ir.Expression, ir.UInt]] = None,
+        localparams: Optional[dict[str, ir.UInt]] = None,
         header: Optional[Lines] = None,
     ):
         self.name = name
 
         input_lines = Lines()
         if is_not_testbench:
-            input_lines += "// Function parameters:"
+            input_lines += (
+                "// Function parameters (only need to be set when start is high):"
+            )
             for input_ in inputs:
                 assert isinstance(input_, str)
                 input_lines += f"input wire signed [31:0] {input_},"
@@ -186,7 +219,7 @@ class Module(ImplementsToLines):
 
         output_lines = Lines()
         if is_not_testbench:
-            output_lines += "output reg _valid, // is high if output is valid"
+            output_lines += "output reg _valid, // is high if output values are valid"
             output_lines.blank()
             output_lines += "output reg _done, // is high if module done outputting"
             output_lines.blank()
@@ -204,10 +237,17 @@ class Module(ImplementsToLines):
             self.body = []
 
         if localparams:
-            assert_typed_dict(localparams, str, ir.UInt)  # type: ignore[misc]
-            self.local_params = Lines()
-            for key, value in localparams.items():
-                self.local_params.concat(LocalParam(key, value).to_lines())
+            if env.get_var(env.Vars.IS_SYSTEM_VERILOG) is not None:
+                self.local_params = Lines("// State variables")
+                self.local_params.concat(
+                    TypeDef("_state_t", list(localparams.keys())).to_lines()
+                )
+                self.local_params += "_state_t _state;"
+            else:
+                self.local_params = Lines("// State variables")
+                for key, value in localparams.items():
+                    self.local_params.concat(LocalParam(key, value).to_lines())
+                self.local_params.concat(Declaration("_state", reg=True).to_lines())
         else:
             self.local_params = Lines()
 
@@ -222,17 +262,21 @@ class Module(ImplementsToLines):
         else:
             lines = Lines()
         lines += f"module {self.name} ("
-        for line in self.input:
-            lines += Indent(1) + line
-        for line in self.output:
-            lines += Indent(1) + line
-        if len(lines) > 1:  # This means there are ports
+        lines.concat(self.input, indent=1)
+        lines.concat(self.output, indent=1)
+
+        if self.input or self.output:  # This means there are ports
             lines[-1] = lines[-1][0:-1]  # removes last comma
+
         lines += ");"
         lines.concat(self.local_params, 1)
         for stmt in self.body:
             lines.concat(stmt.to_lines(), 1)
         lines += "endmodule"
+        lines.blank()
+        lines += "/*"
+        lines.concat(get_mit_license())
+        lines += "*/"
         return lines
 
 
@@ -245,14 +289,15 @@ class Initial(Statement):
 
     def __init__(self, *args, body: Optional[list[Statement]] = None, **kwargs):
         if body:
-            get_typed_list(body, Statement)
+            typed_list(body, Statement)
         self.body = body
         super().__init__(*args, **kwargs)
 
     def to_lines(self):
         lines = Lines("initial begin")
-        for stmt in self.body:
-            lines.concat(stmt.to_lines(), 1)
+        if self.body:
+            for stmt in self.body:
+                lines.concat(stmt.to_lines(), 1)
         lines += "end"
         return lines
 
@@ -285,7 +330,7 @@ class Always(Statement):
         """
         To Verilog
         """
-        lines = Lines(f"always {self.trigger.to_string()} begin")
+        lines = Lines(f"always {self.trigger.verilog()} begin")
         for stmt in self.body:
             lines.concat(stmt.to_lines(), 1)
         lines += "end"
@@ -313,14 +358,14 @@ class Subsitution(Statement):
 
     def __init__(
         self,
-        lvalue: ir.Expression,
+        lvalue: ir.Var,
         rvalue: ir.Expression,
         oper: str,
         *args,
         **kwargs,
     ):
         assert isinstance(rvalue, (ir.Expression)), f"got {type(rvalue)} instead"
-        assert isinstance(lvalue, (ir.Expression))
+        assert isinstance(lvalue, (ir.Expression)), f"{lvalue}"
         self.lvalue = lvalue
         self.rvalue = rvalue
         self.oper = oper
@@ -332,7 +377,7 @@ class Subsitution(Statement):
         Converts to Verilog
         """
         assert isinstance(self.oper, str), "Subclasses need to set self.type"
-        self.literal = f"{self.lvalue} {self.oper} {self.rvalue};"
+        self.literal = f"{self.lvalue.verilog()} {self.oper} {self.rvalue.verilog()};"
         return super().to_lines()
 
 
@@ -341,7 +386,7 @@ class NonBlockingSubsitution(Subsitution):
     <lvalue> <= <rvalue>
     """
 
-    def __init__(self, lvalue: ir.Expression, rvalue: ir.Expression, *args, **kwargs):
+    def __init__(self, lvalue: ir.Var, rvalue: ir.Expression, *args, **kwargs):
         super().__init__(lvalue, rvalue, "<=", *args, **kwargs)
 
 
@@ -350,7 +395,7 @@ class BlockingSub(Subsitution):
     <lvalue> = <rvalue>
     """
 
-    def __init__(self, lvalue: ir.Expression, rvalue: ir.Expression, *args, **kwargs):
+    def __init__(self, lvalue: ir.Var, rvalue: ir.Expression, *args, **kwargs):
         super().__init__(lvalue, rvalue, "=", *args, **kwargs)
 
 
@@ -361,17 +406,17 @@ class Declaration(Statement):
 
     def __init__(
         self,
-        name: ir.Expression | str,
+        name: str,
         *args,
         size: int = 32,
         reg: bool = False,
         signed: bool = False,
         **kwargs,
     ):
-        self.size = size
-        self.reg = reg
-        self.signed = signed
-        self.name = name
+        self.size = typed_strict(size, int)
+        self.reg = typed_strict(reg, bool)
+        self.signed = typed_strict(signed, bool)
+        self.name = typed_strict(name, str)
         super().__init__(*args, **kwargs)
 
     def to_lines(self):
@@ -471,13 +516,13 @@ class IfElse(Statement):
         **kwargs,
     ):
         super().__init__(*args, **kwargs)
-        self.condition = get_typed(condition, ir.Expression)
-        self.then_body = get_typed_list(then_body, Statement)
-        self.else_body = get_typed_list(else_body, Statement)
+        self.condition = typed_strict(condition, ir.Expression)
+        self.then_body = typed_list(then_body, Statement)
+        self.else_body = typed_list(else_body, Statement)
 
     def to_lines(self):
         lines = Lines()
-        lines += f"if ({self.condition.to_string()}) begin"
+        lines += f"if ({self.condition.verilog()}) begin"
         for stmt in self.then_body:
             lines.concat(stmt.to_lines(), indent=1)
         if self.else_body:
@@ -506,13 +551,14 @@ class While(Statement):
         assert isinstance(condition, ir.Expression)
         self.condition = condition
         if body:
-            get_typed_list(body, Statement)
+            typed_list(body, Statement)
         self.body = body
         super().__init__(*args, **kwargs)
 
     def to_lines(self):
         lines = Lines(f"while ({self.condition.to_string()}) begin")
-        for stmt in self.body:
-            lines.concat(stmt.to_lines(), 1)
+        if self.body:
+            for stmt in self.body:
+                lines.concat(stmt.to_lines(), 1)
         lines += "end"
         return lines
