@@ -8,20 +8,20 @@ from __future__ import annotations
 import ast
 import copy
 import io
+import logging
+import warnings
 from dataclasses import dataclass, field
 from types import FunctionType
-from typing import Any, Optional
+from typing import Any, Optional, Union
 
 from python2verilog.api.modes import Modes
 from python2verilog.ir.expressions import ExclusiveVar, State, Var
 from python2verilog.ir.graph import DoneNode
 from python2verilog.ir.instance import Instance
 from python2verilog.ir.signals import ProtocolSignals
-from python2verilog.utils.assertions import assert_typed_dict, get_typed, get_typed_list
 from python2verilog.utils.env import is_debug_mode
 from python2verilog.utils.generics import GenericReprAndStr
-
-DEFAULT_STATE_NAME = "___PYTHON_2_VERILOG_STATE___"
+from python2verilog.utils.typed import guard_dict, typed, typed_list, typed_strict
 
 
 @dataclass
@@ -35,7 +35,7 @@ class Context(GenericReprAndStr):
     name: str = ""
     testbench_suffix: str = "_tb"
 
-    test_cases: list[tuple[int]] = field(default_factory=list)
+    test_cases: list[tuple[int, ...]] = field(default_factory=list)
 
     py_func: Optional[FunctionType] = None
     py_string: Optional[str] = None
@@ -101,12 +101,49 @@ class Context(GenericReprAndStr):
         assert isinstance(self.py_ast, ast.FunctionDef), self
         assert isinstance(self.py_func, FunctionType), self
 
-        def check_list(list_: list):
+        def check_list(list_: list[Any]):
             return isinstance(list_, list) and len(list_) > 0
 
-        assert check_list(self.input_types), f"Input types not inferred for {self.name}"
+        if not check_list(self.input_types):
+
+            def input_mapper(arg: ast.arg) -> type[Any]:
+                """
+                Maps a string annotation id to type
+                """
+                assert arg.annotation, f"{ast.dump(arg)}"
+                assert isinstance(arg.annotation, ast.Name)
+                if arg.annotation.id == "int":
+                    return type(0)
+                raise TypeError(f"{ast.dump(arg)}")
+
+            logging.info(f"Using type hints of {self.name} for input types")
+            input_args: list[ast.arg] = self.py_ast.args.args
+            assert isinstance(input_args, list), f"{ast.dump(self.py_ast)}"
+            self.input_types = list(map(input_mapper, input_args))
+        assert check_list(self.input_types), self
         assert check_list(self.input_vars), self
 
+        if not check_list(self.output_types):
+
+            def output_mapper(arg: ast.Name) -> type[Any]:
+                """
+                Maps a string annotation id to type
+                """
+                if arg.id == "int":
+                    return type(0)
+                raise TypeError(f"{ast.dump(arg)}")
+
+            logging.info(f"Using type hints of {self.name} for return types")
+            output_args: list[ast.arg]
+            if isinstance(self.py_ast.returns, ast.Subscript):
+                assert isinstance(self.py_ast.returns.slice, ast.Tuple)
+                output_args = self.py_ast.returns.slice.elts
+            else:
+                output_args = [self.py_ast.returns]
+            assert isinstance(output_args, list), f"{ast.dump(self.py_ast)}"
+            self.output_types = list(map(output_mapper, output_args))
+            self.default_output_vars()
+
         assert check_list(self.output_types), self
         assert check_list(self.output_vars), self
 
@@ -117,7 +154,7 @@ class Context(GenericReprAndStr):
             assert str(self.entry_state) in self.states, self
 
         for value in self.signals.values():
-            assert get_typed(value, Var)
+            assert typed(value, Var)
 
         return self
 
@@ -195,7 +232,7 @@ class Context(GenericReprAndStr):
 
     @input_vars.setter
     def input_vars(self, other: list[Var]):
-        self._input_vars = get_typed_list(other, Var)
+        self._input_vars = typed_list(other, Var)
 
     @property
     def output_vars(self):
@@ -206,14 +243,16 @@ class Context(GenericReprAndStr):
 
     @output_vars.setter
     def output_vars(self, other: list[Var]):
-        self._output_vars = get_typed_list(other, Var)
+        self._output_vars = typed_list(other, Var)
 
     def default_output_vars(self):
         """
         Sets own output vars to default based on number of output variables
         """
         assert self.output_types and len(self.output_types) > 0
-        self._output_vars = [Var(f"out{i}") for i in range(len(self.output_types))]
+        self._output_vars = [
+            ExclusiveVar(f"out{i}") for i in range(len(self.output_types))
+        ]
 
     @property
     def global_vars(self):
@@ -224,20 +263,20 @@ class Context(GenericReprAndStr):
 
     @global_vars.setter
     def global_vars(self, other: list[Var]):
-        self._global_vars = get_typed_list(other, Var)
+        self._global_vars = typed_list(other, Var)
 
     def add_global_var(self, var: Var):
         """
         Appends global var
         """
-        var = get_typed(var, Var)
+        var = typed_strict(var, Var)
         if (
             var in self._global_vars
             or var in self._input_vars
             or var in self.output_vars
         ):
             return
-        self._global_vars.append(get_typed(var, Var))
+        self._global_vars.append(typed_strict(var, Var))
 
     @property
     def states(self):
@@ -307,6 +346,8 @@ class Context(GenericReprAndStr):
         """
         Create generator instance
         """
+        self.validate()
+
         inst_input_vars: list[Var] = list(
             map(
                 lambda var: ExclusiveVar(f"{name}_{self.name}_{var.py_name}"),
