@@ -10,7 +10,7 @@ import tempfile
 from typing import Iterable, Optional, Union
 
 from python2verilog.utils import env
-from python2verilog.utils.assertions import assert_typed_dict
+from python2verilog.utils.typed import guard_dict
 
 
 def make_cmd(top_level_module: str, files: Iterable[Union[str, os.PathLike[str]]]):
@@ -52,7 +52,7 @@ def _run_cmd_with_fifos(
 
     :return: (stdout, stderr/exception)
     """
-    assert_typed_dict(input_fifos, str, str)  # type: ignore
+    guard_dict(input_fifos, str, str)  # type: ignore
     # pylint: disable=subprocess-popen-preexec-fn
     with subprocess.Popen(
         command,
@@ -90,7 +90,7 @@ def _run_cmd_with_files(
 
     :return: (stdout, stderr/exception)
     """
-    assert_typed_dict(input_files, str, str)  # type: ignore
+    guard_dict(input_files, str, str)  # type: ignore
     # _write_data_to_paths(input_files)
     # pylint: disable=subprocess-popen-preexec-fn
     with subprocess.Popen(
